from typing import Callable

from fastapi import Depends, HTTPException
from pydantic import BaseModel

from . import CRUDGenerator, NOT_FOUND, utils

try:
    from sqlalchemy.orm import Session
    from sqlalchemy.ext.declarative import DeclarativeMeta
    from sqlalchemy.exc import IntegrityError
except ImportError:
    sqlalchemy_installed = False
    Session = None
    DeclarativeMeta = None
else:
    sqlalchemy_installed = True


class SQLAlchemyCRUDRouter(CRUDGenerator):

    def __init__(self, schema: BaseModel, db_model: DeclarativeMeta, db: Session, *args, **kwargs):
        assert sqlalchemy_installed, "SQLAlchemy must be installed to use the SQLAlchemyCRUDRouter."

        self.db_model = db_model
        self.db_func = db
        self._pk: str = db_model.__table__.primary_key.columns.keys()[0]
        self._pk_type: type = utils.get_pk_type(schema, self._pk)

        if 'prefix' not in kwargs:
            kwargs['prefix'] = db_model.__tablename__

        if 'create_schema' not in kwargs:
            kwargs['create_schema'] = self.schema_factory(schema, self._pk)

        super().__init__(schema, *args, **kwargs)

    def _get_all(self) -> Callable:
        def route(db: Session = Depends(self.db_func)):
            return db.query(self.db_model).all()

        return route

    def _get_one(self) -> Callable:
        def route(item_id: self._pk_type, db: Session = Depends(self.db_func)):
            model = db.query(self.db_model).get(item_id)

            if model:
                return model
            else:
                raise NOT_FOUND

        return route

    def _create(self) -> Callable:
        def route(model: self.create_schema, db: Session = Depends(self.db_func)):
            try:
                db_model = self.db_model(**model.dict())
                db.add(db_model)
                db.commit()
                db.refresh(db_model)
                return db_model
            except IntegrityError:
                db.rollback()
                raise HTTPException(422, 'Key already exists')

        return route

    def _update(self) -> Callable:
        def route(item_id: self._pk_type, model: self.schema, db: Session = Depends(self.db_func)):
            db_model = self._get_one()(item_id, db)

            for key, value in model.dict(exclude={self._pk}).items():
                if hasattr(db_model, key):
                    setattr(db_model, key, value)

            db.commit()
            db.refresh(db_model)

            return db_model

        return route

    def _delete_all(self) -> Callable:
        def route(db: Session = Depends(self.db_func)):
            db.query(self.db_model).delete()
            db.commit()

            return self._get_all()(db)

        return route

    def _delete_one(self) -> Callable:
        def route(item_id: self._pk_type, db: Session = Depends(self.db_func)):
            db_model = self._get_one()(item_id, db)
            db.delete(db_model)
            db.commit()

            return db_model

        return route



from . import utils
from ._base import CRUDGenerator, NOT_FOUND

from .mem import MemoryCRUDRouter
from .sqlalchemy import SQLAlchemyCRUDRouter
from .databases import DatabasesCRUDRouter
from .tortoise import TortoiseCRUDRouter



from typing import Callable
from fastapi import Depends, HTTPException
from pydantic import BaseModel

from . import CRUDGenerator, NOT_FOUND, utils

try:
    from sqlalchemy.sql.schema import Table
    from databases.core import Database
except ImportError:
    databases_installed = False
    Table = None
    Database = None
else:
    databases_installed = True


class DatabasesCRUDRouter(CRUDGenerator):

    def __init__(self, schema: BaseModel, table: Table, database: Database, *args, **kwargs):
        assert databases_installed, "Databases and SQLAlchemy must be installed to use the DatabasesCRUDRouter."
        self.table = table
        self.db = database
        self._pk = table.primary_key.columns.values()[0].name
        self._pk_col = self.table.c[self._pk]
        self._pk_type: type = utils.get_pk_type(schema, self._pk)

        if 'prefix' not in kwargs:
            kwargs['prefix'] = table.name

        if 'create_schema' not in kwargs:
            kwargs['create_schema'] = self.schema_factory(schema, self._pk)

        super().__init__(schema, *args, **kwargs)

    def _get_all(self) -> Callable:
        async def route():
            q = self.table.select()
            return await self.db.fetch_all(q)

        return route

    def _get_one(self) -> Callable:
        async def route(item_id: self._pk_type):
            q = self.table.select().where(self._pk_col == item_id)
            model = await self.db.fetch_one(q)

            if model:
                return model
            else:
                raise NOT_FOUND

        return route

    def _create(self) -> Callable:
        async def route(schema: self.create_schema):
            try:
                q = self.table.insert()
                rid = await self.db.execute(query=q, values=schema.dict())
                return {self._pk: rid, **schema.dict()}
            except Exception:
                raise HTTPException(422, 'Key already exists')

        return route

    def _update(self) -> Callable:
        async def route(item_id: self._pk_type, schema: self.schema):
            q = self.table.update().where(self._pk_col == item_id)
            rid = await self.db.execute(query=q, values=schema.dict(exclude={self._pk}))

            if rid:
                return {self._pk: rid, **schema.dict()}
            else:
                raise NOT_FOUND

        return route

    def _delete_all(self) -> Callable:
        async def route():
            q = self.table.delete()
            await self.db.execute(query=q)

            return await self._get_all()()

        return route

    def _delete_one(self) -> Callable:
        async def route(item_id: self._pk_type):
            q = self.table.delete().where(self._pk_col == item_id)

            row = await self._get_one()(item_id)
            rid = await self.db.execute(query=q)

            if rid:
                return row
            else:
                raise NOT_FOUND

        return route



from typing import Callable
from pydantic import BaseModel


from . import CRUDGenerator, NOT_FOUND


class MemoryCRUDRouter(CRUDGenerator):
    
    def __init__(self, schema: BaseModel, *args, **kwargs):
        super(MemoryCRUDRouter, self).__init__(schema, *args, **kwargs)
        self.models = []
        self._id = 0

    def _get_all(self) -> Callable:
        def route():
            return self.models
        return route

    def _get_one(self) -> Callable:
        def route(item_id: int):
            for m in self.models:
                if m.id == item_id:
                    return m

            raise NOT_FOUND

        return route

    def _create(self) -> Callable:
        def route(model: self.schema):
            model.id = self._get_next_id()
            self.models.append(model)
            return model

        return route

    def _update(self) -> Callable:
        def route(item_id: int, model: self.schema):
            for i, m in enumerate(self.models):
                if m.id == item_id:
                    model.id = m.id
                    self.models[i] = model
                    return model

            raise NOT_FOUND
        return route

    def _delete_all(self) -> Callable:
        def route():
            self.models = []
            return self.models

        return route

    def _delete_one(self) -> Callable:
        def route(item_id: int):
            for i, m in enumerate(self.models):
                if m.id == item_id:
                    del self.models[i]
                    return m

            raise NOT_FOUND

        return route

    def _get_next_id(self) -> int:
        id = self._id
        self._id += 1

        return id



from setuptools import setup, find_packages

VERSION = '0.4.1'

setup(
    name='fastapi-crudrouter',
    version=VERSION,
    author='Adam Watkins',
    author_email='cadamrun@gmail.com',
    packages=find_packages(exclude=('tests', 'tests.implementations')),
    url='https://github.com/awtkns/fastapi-crudrouter',
    documentation='https://fastapi-crudrouter.awtkns.com/',
    license='MIT',
    description='A dynamic FastAPI router that automatically creates CRUD routes for your models',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    install_requires=["fastapi"],
    python_requires='>=3.6',
    keywords=['fastapi', 'crud', 'restful', 'routing', 'generator'],
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python",
        "Topic :: Internet",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development",
        "Typing :: Typed",
        "Development Status :: 4 - Beta",
        "Environment :: Web Environment",
        "Framework :: AsyncIO",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Internet :: WWW/HTTP :: HTTP Servers",
        "Topic :: Internet :: WWW/HTTP",
    ]
)




from typing import Callable
from pydantic import BaseModel

from . import CRUDGenerator, NOT_FOUND

try:
    from tortoise import Tortoise
    from tortoise.models import Model
except ImportError:
    tortoise_installed = False
else:
    tortoise_installed = True


class TortoiseCRUDRouter(CRUDGenerator):

    def __init__(self, schema: BaseModel, db_model, *args, **kwargs):
        assert tortoise_installed, "Tortoise ORM must be installed to use the TortoiseCRUDRouter."

        self.db_model = db_model
        self._primary_key: str = db_model.describe()['pk_field']['db_column']

        if 'prefix' not in kwargs:
            # unsure why the name has a "None." appended but I handle it
            kwargs['prefix'] = db_model.describe()['name'].replace('None.', '')

        if 'create_schema' not in kwargs:
            kwargs['create_schema'] = self.schema_factory(schema, self._primary_key)

        super().__init__(schema, *args, **kwargs)

    def _get_all(self) -> Callable:
        async def route():
            models = await self.db_model.all()
            return models

        return route

    def _get_one(self) -> Callable:
        async def route(item_id):
            model = await self.db_model.filter(id=item_id).first()

            if model:
                return model
            else:
                raise NOT_FOUND

        return route

    def _create(self) -> Callable:
        async def route(model: self.create_schema):
            db_model = self.db_model(**model.dict())
            await db_model.save()

            return db_model

        return route

    def _update(self) -> Callable:
        async def route(item_id: int, model: self.create_schema):
            await self.db_model.filter(id=item_id).update(**model.dict(exclude_unset=True))
            return await self._get_one()(item_id)

        return route

    def _delete_all(self) -> Callable:
        async def route():
            await self.db_model.all().delete()
            return await self._get_all()()

        return route

    def _delete_one(self) -> Callable:
        async def route(item_id: int):
            model = await self._get_one()(item_id)
            await self.db_model.filter(id=item_id).delete()

            return model

        return route



from typing import List, Optional, Callable

from starlette.routing import Route
from fastapi import APIRouter, HTTPException
from pydantic import BaseModel, create_model

models = []

NOT_FOUND = HTTPException(404, 'Item not found')


class CRUDGenerator(APIRouter):
    schema: BaseModel = None
    _base_path: str = '/'

    def __init__(
        self,
        schema: BaseModel,
        create_schema: BaseModel = None,
        prefix: str = None,
        get_all_route: bool = True,
        get_one_route: bool = True,
        create_route: bool = True,
        update_route: bool = True,
        delete_one_route: bool = True,
        delete_all_route: bool = True,
        *args,
        **kwargs
    ):

        self.schema = schema
        self.create_schema = create_schema if create_schema else self.schema_factory(self.schema)

        prefix = self._base_path + (self.schema.__name__.lower() if not prefix else prefix).strip('/')
        super().__init__(prefix=prefix, tags=[prefix.strip('/').capitalize()], *args, **kwargs)

        if get_all_route:
            super().add_api_route('', self._get_all(), methods=['GET'], response_model=Optional[List[self.schema]], summary='Get All')

        if create_route:
            super().add_api_route('', self._create(), methods=['POST'], response_model=self.schema, summary='Create One')

        if delete_all_route:
            super().add_api_route('', self._delete_all(), methods=['DELETE'], response_model=Optional[List[self.schema]], summary='Delete All')

        if get_one_route:
            super().add_api_route('/{item_id}', self._get_one(), methods=['GET'], response_model=self.schema, summary='Get One')

        if update_route:
            super().add_api_route('/{item_id}', self._update(), methods=['PUT'], response_model=self.schema, summary='Update One')

        if delete_one_route:
            super().add_api_route('/{item_id}', self._delete_one(), methods=['DELETE'], response_model=self.schema, summary='Delete One')

    def api_route(self, path: str, *args, **kwargs):
        """ Overrides and exiting route if it exists"""
        methods = kwargs['methods'] if 'methods' in kwargs else ['GET']
        self.remove_api_route(path, methods)
        return super().api_route(path, *args, **kwargs)

    def get(self, path, *args, **kwargs):
        self.remove_api_route(path, ['Get'])
        return super().get(path, *args, **kwargs)

    def post(self, path, *args, **kwargs):
        self.remove_api_route(path, ['POST'])
        return super().post(path, *args, **kwargs)

    def put(self, path, *args, **kwargs):
        self.remove_api_route(path, ['PUT'])
        return super().put(path, *args, **kwargs)

    def delete(self, path, *args, **kwargs):
        self.remove_api_route(path, ['DELETE'])
        return super().delete(path, *args, **kwargs)

    def remove_api_route(self, path: str, methods: List[str]):
        methods = set(methods)

        for r in self.routes:
            if r.path == f'{self.prefix}{path}' and r.methods == methods:
                self.routes.remove(r)

    def _get_all(self, *args, **kwargs) -> Callable:
        raise NotImplementedError

    def _get_one(self, *args, **kwargs) -> Callable:
        raise NotImplementedError

    def _create(self, *args, **kwargs) -> Callable:
        raise NotImplementedError

    def _update(self, *args, **kwargs) -> Callable:
        raise NotImplementedError

    def _delete_one(self, *args, **kwargs) -> Callable:
        raise NotImplementedError

    def _delete_all(self, *args, **kwargs) -> Callable:
        raise NotImplementedError

    @staticmethod
    def get_routes() -> list:
        return [
            'get_all', 'create', 'delete_all', 'get_one', 'update', 'delete_one'
        ]

    @staticmethod
    def schema_factory(schema_cls: BaseModel, pk_field_name: str = 'id'):
        """
        Is used to create a CreateSchema which does not contain pk
        """

        fields = {f.name: (f.type_, ...) for f in schema_cls.__fields__.values() if f.name != pk_field_name}

        name = schema_cls.__name__ + 'Create'
        schema = create_model(name, **fields)
        return schema
